#!/usr/bin/env python
"""
This is the launcher for ctznOSX/Titan
"""

import logging
from sys import argv,exit
from itertools import chain
from config import ctznConfig
from titantools.system import hw_serial as get_device_serial
from socket import gethostname
from time import time, strftime, gmtime
from collections import namedtuple
from subprocess import Popen, PIPE
from os import listdir,walk,path,environ
from os.path import dirname, realpath, isfile, join, splitext, basename

# Get ctznOSX Env and Config
CTZNOSX_PATH = (environ.get('CTZNOSX_PATH') or '/var/lib/ctznosx/')
CTZNOSX_CONFIG = join('/etc/', 'ctznosx.conf')

# Config
CONFIG = ctznConfig( CTZNOSX_CONFIG, CTZNOSX_PATH )

# Configurations
logging.basicConfig(format='%(message)s', level=logging.INFO)

# Constants
CURRENT_DIR = dirname(realpath(__file__))

# Base Modules Dir
MODULES_DIR = CONFIG['main']['monitorstore']

# Log Directory
LOG_DIR = CONFIG['main']['logstore']

# Log Directory
DATASTORE = CONFIG['main']['datastore']

# Report Directory
REPORT_DIR = CONFIG['main']['reportstore']

# Get Hostname
HOSTNAME = gethostname()

# Get Serial Number
SERIALNUMBER = get_device_serial()

# Get Runtime
DATE = strftime("%Y-%m-%dT%H:%M:%S%z", gmtime())

# Define Module Packs
MODULE_PACKS = [
    join(MODULES_DIR, mod_pack) for mod_pack in listdir(MODULES_DIR)
]

# Define all of our modules
MODULES = []
for path in MODULE_PACKS:
    for root,dirs,files in walk(path):
        for f in (f for f in files if f not in [".gitkeep", "README", "schema.py", "config.json"]):
            MODULES.append( join(root,f) )

# Types
CtznLanguage = namedtuple("CtznLanguage", "supported_extensions execution_string")

PYTHON_LANGUAGE = CtznLanguage(
    supported_extensions = [".py", ".pyc"],
    execution_string = "python",
)

RUBY_LANGUAGE = CtznLanguage(
    supported_extensions = [".rb"],
    execution_string = "ruby"
)

BASH_LANGUAGE = CtznLanguage(
    supported_extensions = [".bash", ".sh"],
    execution_string = "/bin/bash"
)

PHP_LANGUAGE = CtznLanguage(
    supported_extensions = [".php"],
    execution_string = "php"
)

PERL_LANGUAGE = CtznLanguage(
    supported_extensions = [".pl"],
    execution_string = "perl"
)

SUPPORTED_LANGUAGES = [
    PYTHON_LANGUAGE,
    RUBY_LANGUAGE,
    BASH_LANGUAGE,
    PHP_LANGUAGE,
    PERL_LANGUAGE,
]

# Set Debugging
testing_enabled = False

if "--verbose" in argv[1:]:
    testing_enabled = True

# Functions
def log_line(log_name, line):
    """log_line accepts a line a returns a properly formatted log line"""
    return "%s %s by[%s]: %s" % (
        DATE,
        HOSTNAME,
        log_name,
        line,
    )

# Spawn module, capture output
def spawn_module(module, current_lang, mod_name):
    """spawn_module executes an individual Titan module"""
    log_file = join(LOG_DIR, mod_name + ".log")
    
    if testing_enabled:
        logging_passthru = '--log'
    else:
        logging_passthru = ''

    command = list(chain(
        ["/usr/bin/sudo"], #/usr/bin/sudo -u _ctznosx
        current_lang.execution_string.split(" "),
        [module],
        [DATASTORE],
        [logging_passthru]
    ))
    
    execution = Popen(command, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    stdout = execution.stdout.readlines()
    stderr = execution.stderr.readlines()

    file_handler = open(log_file, "a")

    for stdout_line in stdout:
        if testing_enabled:
            print log_line(mod_name, stdout_line)

        file_handler.write(log_line(mod_name, stdout_line))

    for stderr_line in stderr:
        if testing_enabled:
            print log_line(mod_name, stderr_line)

        file_handler.write(log_line(mod_name, stderr_line))

# Detect Modules
def launch_modules():
    """launch_modules launches Titan's executable modules"""
    for module in MODULES:
        current_lang = None
        mod_name, ext = splitext(basename(module))

        for language in SUPPORTED_LANGUAGES:
            if ext in language.supported_extensions:
                current_lang = language
                break

        if current_lang is not None and isinstance(current_lang, CtznLanguage):
            if testing_enabled:
                print log_line("ctznOSX", "Found Module: %s/%s, Lang: %s" % (basename(dirname(module)),mod_name, current_lang.execution_string))
            
            spawn_module(module, current_lang, mod_name)

# Default run method
def run():
    # Record Start Time
    start = time()

    # Launch Modules
    launch_modules()

    # Record End Time
    end = time()
    logging.info("Execution took %s seconds.", str(round(end - start, 4)))   

# For Mother Import
if __name__ == "__main__":
    run()



import os, sys
from ctznosx import __version__ as version
from distutils.core import setup
from setuptools.command.install import install
from titantools.system import shell_out

long_description = """
ctznOSX is used by system administrators and
security professionals to profile and record details
relating to organizational wide Mac deployments.
"""

install_requires = ['titantools']

if sys.version_info < (2, 7):
    install_requires += ['argparse']

""" ctznOSX Installer Helper """
class CtznosxInstaller(install):
    """ Preinstall hooks """
    def preinstall(self):
        # Add custom pre-install stuff
        print "Adding ctznOSX User"
        add_user_script = """
dscl . -create /Users/_ctznosx
dscl . -create /Users/_ctznosx RealName "ctznOSX Management User"
dscl . -create /Users/_ctznosx UniqueID 403  # Use something between 100 and 500 to hide the user
dscl . -create /Users/_ctznosx PrimaryGroupID 20
dscl . -create /Users/_ctznosx UserShell /sbin/nologin
dseditgroup -o edit -a _ctznosx -t user admin
dseditgroup -o edit -a _ctznosx -t user wheel"""
        shell_out(add_user_script)
    
    """ Postinstall hooks """
    def postinstall(self):
        # Post install script
        print "Updating permissions for ctznOSX"
        shell_out("chown -R _ctznosx /var/lib/ctznosx")
        shell_out("chgrp -R wheel /var/lib/ctznosx")
        shell_out("chmod -R 0633 /var/lib/ctznosx/logs")
        shell_out("chmod -R 0666 /var/lib/ctznosx/ctznosx.db")
        shell_out("sudo launchctl load /Library/LaunchDaemons/plist/com.ctznosx.runner.plist")
        shell_out("sudo launchctl load /Library/LaunchDaemons/plist/com.ctznosx.watcher.plist")

    """ Runtime hooks """
    def run(self):

        self.preinstall()
        # Call the super run
        install.run(self)
        self.postinstall()

setup(
    name='ctznosx',
    version=version,
    description='ctznOSX: OSX Secure Device Management',
    long_description=long_description.strip(),
    author='Mike Mackintosh',
    author_email='mike@shutterstock.com',
    license='MIT',
    url='http://ctznOSX.com',
    packages=['ctznosx'],
    data_files=[('/etc', ['ctznosx.conf']), 
                ('/var/lib/ctznosx/', ['README.md', 'ctznosx.db']), 
                ('/var/lib/ctznosx/monitors/', ['monitors/README.md']), 
                ('/var/lib/ctznosx/logs/', []),
                ('/Library/LaunchDaemons/', ['plist/com.ctznosx.runner.plist', 'plist/com.ctznosx.watcher.plist']),
                ('/var/lib/ctznosx/reports/', [])
               ],
    package_data={'': ['ctznosx.db', 'ctznosx.conf', 'README.md']},
    include_package_data=True,
    scripts=['scripts/ctznosx', 'scripts/ctznosx-watcher'],
    install_requires=install_requires,
    classifiers=[
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
    ],
    cmdclass={
        'install': CtznosxInstaller,
    },
)



#!/usr/bin/env python
#
# Todo: Read config file, use target as test for urlopen2
# Todo: Check to make sure we don't duplicate queue files
#

import sys
import logging
import hashlib
from sys import exit
import urllib2, urllib, httplib, json
from os import path,walk,remove,environ
from os.path import dirname, realpath, isfile, join, splitext, basename
from binascii import b2a_hex,hexlify
from config import ctznConfig
from titantools.orm import TiORM as ctznORM
from titantools.system import shell_out,hw_serial as get_device_serial
from zlib import compress
from time import sleep,strftime,strptime,gmtime,mktime

# Get ctznOSX Env and Config
CTZNOSX_PATH = (environ.get('CTZNOSX_PATH') or '/var/lib/ctznosx/')
CTZNOSX_CONFIG = join('/etc/', 'ctznosx.conf')

# Config
CONFIG = ctznConfig( CTZNOSX_CONFIG, CTZNOSX_PATH )

# Set datastore
DATASTORE = CONFIG['main']['datastore']


# Load ORM 
ORM = ctznORM(DATASTORE)

# Load default schema
ORM.initialize_table('watcher', {u'date': {u'nullable': False, u'type': u'text'}, 
                u'utime': {u'type': u'integer'}, 
                u'module': {u'type': u'text'}, 
                u'status': {u'type': u'integer'}})

# Device Serial
DEVICEID = shell_out("ioreg -c IOPlatformExpertDevice |head -30 |grep IOPlatformSerialNumber | awk '{print $4}'")[1:-2]

# Get reporting target
REPORTING_TARGET = CONFIG['reporting']['target']
if REPORTING_TARGET is None or REPORTING_TARGET == "":
  print "[!] No place to report to"
  exit()


# Check if Watcher is enabled
if CONFIG['watcher']['enabled'] is "false":
  print "[!] Watcher is disabled"
  exit()

# Configure Logging
logging.basicConfig(format='%(message)s', level=logging.INFO)

def generate_reports():

  # Report Timestamp
  unix_time = int(mktime(gmtime()))
  exec_time = strftime("%a, %d %b %Y %H:%M:%S-%Z", gmtime())

  # Query all tables
  all_tables = ORM.select('sqlite_master', 'name',  "type = 'table' and name != 'watcher'")

  # How many wins do we need
  passes_needed = 0
  passes_had = 0

  # Loop through tables
  for table in [table for table in all_tables]:
    # Last success
    last_success = ORM.raw_sql('SELECT * FROM watcher WHERE module="%s" ORDER BY date DESC LIMIT 1' % (table['name']))

    # Get table data
    if len(last_success[0]) is 0:
      # Logging is cool
      temp_utime = unix_time
      ORM.raw_sql("INSERT INTO watcher (date,status,utime,module) VALUES ('%s', 0,'%s', '%s')" % (exec_time, temp_utime, table['name']))
      logging.info("Collecting data for [%s] since ever" % (table['name']))
      results = ORM.select(table['name'], '*')
    else:
      # Logging is cool
      temp_utime = int(last_success[0][4])
      logging.info("Collecting data for [%s] since %s" % (table['name'],last_success[0][1]))
      results = ORM.select(table['name'], '*', 'unixtime > %d' % temp_utime)

    if results is None or len(results) == 0:
      ORM.raw_sql("UPDATE watcher SET status=0, utime = '%d' WHERE module = '%s'" % (temp_utime, table['name']))
      continue

    # Dump table to json
    table_json = json.dumps(results)

    # Compress the data
    compressed = compress(table_json)

    # Digest
    content_digest = hashlib.sha256(compressed).hexdigest()

    # Send the table data upstream
    target = "%s/%s" % (REPORTING_TARGET, DEVICEID)
    logging.info("\tSending request to '%s'" % target)
    code, response = send_request(target, {'serial': DEVICEID, 'digest': content_digest, 'stream': compressed})
    logging.info("\tResponse: [%d]" % (code))

    if code == 202:
      passes_had += 1
      ORM.raw_sql("UPDATE watcher SET status=1, utime = %d WHERE module = '%s'" % (temp_utime, table['name']))
    else:
      ORM.raw_sql("UPDATE watcher SET status=0, utime = %d WHERE module = '%s'" % (temp_utime, table['name']))

  print ORM.select('watcher')
  exit()


# Send the request
def send_request( target, data):

  try:
    request = urllib2.Request(target, urllib.urlencode(data) )
    opener = urllib2.build_opener()
    response = opener.open(request)
    response_object = response.getcode(), response

  except urllib2.HTTPError, e:
    if e.code == 307:
      for line in str(e.headers).splitlines():
        if "Location" in line:
          new_target = line.split(": ", 1)[1]
          response_object = send_request( new_target, data )
    else:
      response_object = e.code, e.read()

  except urllib2.URLError, e:
    response_object = 0, 'Connection Refused'

  return response_object


# Run the meat and potatoes
def run():

  # Check if internet is available
  try_count = 0

  # Testing
  generate_reports()
  exit()
  # End testing

  while True:

    try:
      is_connected = urllib2.urlopen( REPORTING_TARGET )

    except urllib2.URLError:

      # Seconds for timeout
      seconds = 3**try_count
      
      # Exceeded max timeout
      if seconds >= 3600:
        exit()

      try_count += 1
      logging.info("Watcher did not detect a connection, retrying in %d seconds" % seconds)

      # Sleep
      sleep(seconds)

    else:
      logging.info("Watcher detected a connection")
      generate_reports()

if __name__ == "__main__":
  run()
